@@ -34,6 +34,10 @@ from mutmut import (
     MutationID,
     SKIPPED,
 )
+from mutmut.cache import (
+    create_html_report,
+    cached_hash_of_tests,
+)
 from mutmut.cache import register_mutants, update_mutant_status, \
     print_result_cache, cached_mutation_status, hash_of_tests, \
     filename_and_mutation_id_from_pk, cached_test_time, set_cached_test_time, \
@@ -157,7 +161,7 @@ class Config(object):
                  baseline_time_elapsed, test_time_multiplier, test_time_base,
                  backup, dict_synonyms, total, using_testmon, cache_only,
                  tests_dirs, hash_of_tests, pre_mutation, post_mutation,
-                 coverage_data):
+                 coverage_data, paths_to_mutate):
         self.swallow_output = swallow_output
         self.test_command = test_command
         self.covered_lines_by_filename = covered_lines_by_filename
@@ -174,6 +178,7 @@ class Config(object):
         self.post_mutation = post_mutation
         self.pre_mutation = pre_mutation
         self.coverage_data = coverage_data
+        self.paths_to_mutate = paths_to_mutate
 
 
 class Progress(object):
@@ -267,7 +272,7 @@ def main(command, argument, argument2, paths_to_mutate, backup, runner, tests_di
     if use_coverage and use_patch_file:
         raise click.BadArgumentUsage("You can't combine --use-coverage and --use-patch")
 
-    valid_commands = ['run', 'results', 'apply', 'show', 'junitxml']
+    valid_commands = ['run', 'results', 'apply', 'show', 'junitxml', 'html']
     if command not in valid_commands:
         raise click.BadArgumentUsage('{} is not a valid command, must be one of {}'.format(command, ', '.join(valid_commands)))
 
@@ -303,6 +308,10 @@ def main(command, argument, argument2, paths_to_mutate, backup, runner, tests_di
         print_result_cache_junitxml(dict_synonyms, suspicious_policy, untested_policy)
         return 0
 
+    if command == 'html':
+        create_html_report(dict_synonyms)
+        return 0
+
     if command == 'apply':
         do_apply(argument, dict_synonyms, backup)
         return 0
@@ -324,6 +333,7 @@ def main(command, argument, argument2, paths_to_mutate, backup, runner, tests_di
         for pt in tests_dir.split(':'):
             tests_dirs.extend(glob(p + '/**/' + pt, recursive=True))
     del tests_dir
+    current_hash_of_tests = hash_of_tests(tests_dirs)
 
     os.environ['PYTHONDONTWRITEBYTECODE'] = '1'  # stop python from creating .pyc files
 
@@ -351,7 +361,8 @@ Legend for output:
     baseline_time_elapsed = time_test_suite(
         swallow_output=not swallow_output,
         test_command=runner,
-        using_testmon=using_testmon
+        using_testmon=using_testmon,
+        current_hash_of_tests=current_hash_of_tests,
     )
 
     if hasattr(mutmut_config, 'init'):
@@ -393,11 +404,12 @@ Legend for output:
         using_testmon=using_testmon,
         cache_only=cache_only,
         tests_dirs=tests_dirs,
-        hash_of_tests=hash_of_tests(tests_dirs),
+        hash_of_tests=current_hash_of_tests,
         test_time_multiplier=test_time_multiplier,
         test_time_base=test_time_base,
         pre_mutation=pre_mutation,
         post_mutation=post_mutation,
+        paths_to_mutate=paths_to_mutate,
     )
 
     if argument is None:
@@ -406,8 +418,15 @@ Legend for output:
                 update_line_numbers(filename)
                 add_mutations_by_file(mutations_by_file, filename, dict_synonyms, config)
     else:
-        filename, mutation_id = filename_and_mutation_id_from_pk(int(argument))
-        mutations_by_file[filename] = [mutation_id]
+        try:
+            filename, mutation_id = filename_and_mutation_id_from_pk(int(argument))
+            mutations_by_file[filename] = [mutation_id]
+        except ValueError:
+            filename = argument
+            if not os.path.exists(filename):
+                raise click.BadArgumentUsage('The run command takes either an integer that is the mutation id or a path to a file to mutate')
+            update_line_numbers(filename)
+            add_mutations_by_file(mutations_by_file, filename, dict_synonyms, config)
 
     config.total = sum(len(mutations) for mutations in mutations_by_file.values())
 
@@ -522,6 +541,54 @@ def tests_pass(config: Config, callback) -> bool:
     if config.using_testmon:
         copy('.testmondata-initial', '.testmondata')
 
+    use_special_case = False
+
+    # Special case for hammett! We can do in-process test running which is much faster
+    hammett_prefix = 'python -m hammett '
+    if use_special_case and config.test_command.startswith(hammett_prefix):
+        # noinspection PyUnresolvedReferences
+        from hammett import main_cli
+        modules_before = set(sys.modules.keys())
+
+        # set up timeout
+        import _thread
+        from threading import (
+            Timer,
+            current_thread,
+            main_thread,
+        )
+
+        timed_out = False
+
+        def timeout():
+            _thread.interrupt_main()
+            nonlocal timed_out
+            timed_out = True
+
+        assert current_thread() is main_thread()
+        timer = Timer(config.baseline_time_elapsed * 10, timeout)
+        timer.daemon = True
+        timer.start()
+
+        # Run tests
+        try:
+            returncode = main_cli(shlex.split(config.test_command[len(hammett_prefix):]))
+            timer.cancel()
+        except KeyboardInterrupt:
+            timer.cancel()
+            if timed_out:
+                raise TimeoutError('In process tests timed out')
+            raise
+
+        modules_to_force_unload = {x.partition(os.sep)[0] for x in config.paths_to_mutate}
+
+        for module_name in list(sorted(sys.modules.keys(), reverse=True)):
+            if module_name not in modules_before:
+                if any(module_name.startswith(x) for x in modules_to_force_unload) or module_name.startswith('tests') or module_name.startswith('django'):
+                    del sys.modules[module_name]
+
+        return returncode == 0
+
     returncode = popen_streaming_output(config.test_command, callback, timeout=config.baseline_time_elapsed * 10)
     return returncode == 0 or (config.using_testmon and returncode == 5)
 
@@ -539,10 +606,11 @@ def run_mutation(config: Config, filename: str, mutation_id: MutationID, callbac
 
     cached_status = cached_mutation_status(filename, mutation_id, config.hash_of_tests)
 
-    if cached_status != UNTESTED:
+    if cached_status != UNTESTED and config.total != 1:
         return cached_status
 
     if hasattr(mutmut_config, 'pre_mutation'):
+        context.current_line_index = context.mutation_id.line_number
         mutmut_config.pre_mutation(context=context)
         config = context.config
         if context.skip:
@@ -655,7 +723,7 @@ def read_patch_data(patch_file_path):
     }
 
 
-def time_test_suite(swallow_output, test_command, using_testmon):
+def time_test_suite(swallow_output, test_command, using_testmon, current_hash_of_tests):
     """Execute a test suite specified by ``test_command`` and record
     the time it took to execute the test suite as a floating point number
 
@@ -673,7 +741,7 @@ def time_test_suite(swallow_output, test_command, using_testmon):
     :rtype: float
     """
     cached_time = cached_test_time()
-    if cached_time is not None:
+    if cached_time is not None and current_hash_of_tests == cached_hash_of_tests():
         print('1. Using cached time for baseline tests, to run baseline again delete the cache file')
         return cached_time
 
@@ -697,7 +765,7 @@ def time_test_suite(swallow_output, test_command, using_testmon):
 
     print('Done')
 
-    set_cached_test_time(baseline_time_elapsed)
+    set_cached_test_time(baseline_time_elapsed, current_hash_of_tests)
 
     return baseline_time_elapsed
 
